;compiler: avrasm

	
;************************************************************************************************
;	8-bit HD44780 compatibile LCD routines
;************************************************************************************************
;------------------------------------------------------------------------------------------------
;	send R16 byte, wait ~42us
;------------------------------------------------------------------------------------------------
LCDwrByte:	pusht	R16
	in	R17,pROE
	sbi	pROE,ROE	;SRAM OE=1
	in	R18,ddrc
	out	portc,R16	;put byte on port
	outi	ddrc,0xFF	;set it as output
	adiw	R30,0
	sbi	pLCK,LCK
	adiw	R30,0	
	cbi	pLCK,LCK	;clock
	ldi	R16,41*XTAL/(7*1000000)	;delay
	adiw	R30,0
	adiw	R30,0
	dbnz	R16,PC-1-1-1
	out	ddrc,R18	;set port back
	tst	R18
	brne	PC+1+2
	outi	portc,0xFF	;pull-ups if configured as inputs
	out	pROE,R17
	popt	R16
	ret
	
;------------------------------------------------------------------------------------------------
;	send R16 instruction
;------------------------------------------------------------------------------------------------
LCDwrInstr:	cbi	pLRS,LRS	;RS=0
	rjmp	LCDwrByte	;send byte

;------------------------------------------------------------------------------------------------
;	send R16 data
;------------------------------------------------------------------------------------------------
LCDwrData:	sbi	pLRS,LRS	;RS=1
	rjmp	LCDwrByte	;send byte

;------------------------------------------------------------------------------------------------
;	clear entire DDRAM
;------------------------------------------------------------------------------------------------
LCDclrFull:	ldi	R16,0|128
	rcall	LCDwrInstr	;reset cursor
	;write spaces
	ldi	R18,0x80
LCDclrFullL1:	ldi	R16,' '
	rcall	LCDwrData
	dbnz	R18,LCDclrFullL1
	ldi	R16,0|128
	rcall	LCDwrInstr	;reset cursor	
	ret
	
;------------------------------------------------------------------------------------------------
;	clear 2x20 chars (write ' ')
;------------------------------------------------------------------------------------------------
LCDclrAll:	;write spaces line 1
	ldi	R16,0|128
	rcall	LCDwrInstr	;reset cursor
	ldi	R17,20
LCDclrAllL1:	ldi	R16,' '
	rcall	LCDwrData
	dbnz	R17,LCDclrAllL1
	;write spaces line 2
	ldi	R16,0x40|128
	rcall	LCDwrInstr	;reset cursor
	ldi	R17,20
LCDclrAllL2:	ldi	R16,' '
	rcall	LCDwrData
	dbnz	R17,LCDclrAllL2
	ldi	R16,0|128
	rcall	LCDwrInstr	;reset cursor	
	ret	

;------------------------------------------------------------------------------------------------
;	clear first line
;------------------------------------------------------------------------------------------------
LCDclrL1:	ldi	R16,0|128
	rcall	LCDwrInstr	;reset cursor
	ldi	R17,20
LCDclrL1L1:	ldi	R16,' '
	rcall	LCDwrData
	dbnz	R17,LCDclrL1L1
	ldi	R16,0|128
	rcall	LCDwrInstr	;reset cursor
	ret
	
;------------------------------------------------------------------------------------------------
;	clear second line
;------------------------------------------------------------------------------------------------
LCDclrL2:	ldi	R16,0x40|128
	rcall	LCDwrInstr	;reset cursor
	ldi	R17,20
LCDclrL2L1:	ldi	R16,' '
	rcall	LCDwrData
	dbnz	R17,LCDclrL2L1
	ldi	R16,0x40|128
	rcall	LCDwrInstr	;reset cursor
	ret		
	
;------------------------------------------------------------------------------------------------
;	set pos to line 1
;------------------------------------------------------------------------------------------------
LCDposL1:	push	R16
	ldi	R16,0x00|128
	rcall	LCDwrInstr
	pop	R16
	ret

;------------------------------------------------------------------------------------------------
;	set pos to line 2
;------------------------------------------------------------------------------------------------	
LCDposL2:	push	R16
	ldi	R16,0x40|128
	rcall	LCDwrInstr
	pop	R16
	ret	

;------------------------------------------------------------------------------------------------
;	init LCD as 8-bit 2x20 characters
;------------------------------------------------------------------------------------------------
LCDinit:	ldi	R16,0b00001100
	rcall	LCDwrInstr	;display on
	ldi	R16,0b00111000
	rcall	LCDwrInstr	;2lines, 8-bit, 5x8 char
	rcall	GenerovatZnaky	;generate VUs
	rcall	LCDclrFull	;clear all DDRAM
	ret
	
;------------------------------------------------------------------------------------------------
;	send Z pointed FLASH string (0 terminated)
;------------------------------------------------------------------------------------------------
LCDwrStrFlash:	lpm	R16,Z+
	cpi	R16,0x0D
	brne	PC+1+1+1+1
	ldi	R16,0x40|128
	rcall	LCDwrInstr	;goto second line
	rjmp	LCDwrStrFlash
	cpi	R16,0
	brne	PC+1+1
	ret		;leave if R16=0
	rcall	LCDwrData
	rjmp	LCDwrStrFlash
	
;------------------------------------------------------------------------------------------------
;	send Z pointed FLASH string (0 terminated), usage: LwrFS TText
;------------------------------------------------------------------------------------------------	
	.macro	LwrFS
	ldi	R31,high(@0*2)
	ldi	R30,low(@0*2)
	rcall	LCDwrStrFlash
	.endm
	
;------------------------------------------------------------------------------------------------
;	send Z pointed RAM string
;------------------------------------------------------------------------------------------------
LCDwrStrRam:	ld	R16,Z+
	rcall	LCDwrData
	dbnz	R17,LCDwrStrRam
	ret
	
	;vygeneruje znaky pro bargrafy do CGRAM
GenerovatZnaky:	ldi	R19,0b01000000
	ldi	R17,8
	ldi	R18,1
GenZnaky1:	;cbi	portb,1
	mov	R16,R19
	inc	R19
	rcall	LCDwrInstr
	;call	ByteToLCD
	;sbi	portb,1
	ldi	R16,0b00000000
	cp	R18,R17
	brcs	NeBlack
	ldi	R16,0b00011111
NeBlack:	;call	ByteToLCD
	rcall	LCDwrData
	inc	R18
	cpi	R18,9
	brne	DalsiBarCh
	ldi	R18,1
	dec	R17
DalsiBarCh:	cpi	R19,0b10000000
	brne	GenZnaky1
	;cbi	portb,1
	ret


;------------------------------------------------------------------------------------------------
;	print 2 digits of R16 number, convert leading zeroes to ' '
;------------------------------------------------------------------------------------------------
LCDwrBCD2:	call	bin2BCD8u2
	push	R16
	mov	R16,R17
	clt
	rcall	LCDwrBCD3sc
	pop	R16
	set
	rcall	LCDwrBCD3sc
	ret

;------------------------------------------------------------------------------------------------
;	print 3 digits of R16 number, convert leading zeroes to ' '
;------------------------------------------------------------------------------------------------	
LCDwrBCD3:	call	bin2BCD8u3
	push	R16
	push	R17
	clt
	mov	R16,R18
	rcall	LCDwrBCD3sc
	pop	R16
	rcall	LCDwrBCD3sc
	set
	pop	R16
	rcall	LCDwrBCD3sc
	ret
LCDwrBCD3sc:	cpi	R16,0
	breq	PC+1+1
	set
	brts	PC+1+1
	ldi	R16,' '-'0'
	subi	R16,-'0'
	rcall	LCDwrData
	ret
	

;------------------------------------------------------------------------------------------------
;	print 3 digits of R16 number, skip leading zeroes
;------------------------------------------------------------------------------------------------
LCDwrBCD3sz:	call	bin2BCD8u3
	pushw	R16
	clt
	mov	R16,R18
	rcall	LCDwrBCD3szc
	pop	R16	
	rcall	LCDwrBCD3szc
	set
	pop	R16
	rcall	LCDwrBCD3szc
	ret	
LCDwrBCD3szc:	brts	PC+1+1+1+1
	cpi	R16,0
	breq	PC+1+1+1+1
	set
	subi	R16,-'0'
	rcall	LCDwrData
	ret


;------------------------------------------------------------------------------------------------
;	print 8 digits of R16:R17:R18 number, skip leading zeroes
;------------------------------------------------------------------------------------------------
LCDwrBCD8sz:	call	bin2BCD24p8	;convert to BCD
	clt
	mov	R16,R15
	swap	R16
	rcall	LCDwrBCDsz
	mov	R16,R15
	rcall	LCDwrBCDsz
	mov	R16,R14
	swap	R16
	rcall	LCDwrBCDsz
	mov	R16,R14
	rcall	LCDwrBCDsz
	mov	R16,R13
	swap	R16
	rcall	LCDwrBCDsz
	mov	R16,R13
	rcall	LCDwrBCDsz
	mov	R16,R12
	swap	R16
	rcall	LCDwrBCDsz
	set
	mov	R16,R12
	rcall	LCDwrBCDsz
	ret
LCDwrBCDsz:	andi	R16,0x0F
	brts	PC+1+1+1
	brne	PC+1+1
	ret
	set
	subi	R16,-'0'
	rcall	LCDwrData
	ret

;------------------------------------------------------------------------------------------------
;	send Z pointed RAM string, usage: LwrRS len TText
;------------------------------------------------------------------------------------------------	
	.macro	LwrRS
	ldi	R17,@0
	ldi	R31,high(@1)
	ldi	R30,low(@1)
	rcall	LCDwrStrRam
	.endm	
	

;------------------------------------------------------------------------------------------------
;	Draw MOD play status
;------------------------------------------------------------------------------------------------
LCDwrMODln:	ldi	R16,0x40|128
	rcall	LCDwrInstr	;second line
	
	;print line
	lds	R16,MODpln
	dec	R16
	rcall	LCDwrBCD2
	ldi	R16,'/'
	rcall	LCDwrData
	
	;print pattern
	lds	R16,MODpat
	inc	R16
	rcall	LCDwrBCD3
	ldi	R16,'/'
	rcall	LCDwrData
	
	;print total pattern count
	lds	R16,MODpatn
	rcall	LCDwrBCD3
	ret
	
	
;------------------------------------------------------------------------------------------------
;	Draw MOD vu's
;------------------------------------------------------------------------------------------------
LCDwrMODvu:	.if	DRAWVU==1
	ldi	R16,(0x40+20)|128
	lds	R17,MODchnn
	cpi	R17,10+1
	brlo	PC+1+1
	ldi	R17,10
	sub	R16,R17
	push	R17
		
	rcall	LCDwrInstr	;second line
	ldiw	R28,MODbars
	;ldi	R24,4
	;lds	R24,MODchnn
	pop	R24
LCDwrMODlnBGL1:	ld	R16,Y	;load peak value
	cpi	R16,0
	brge	PC+1+1
	ldi	R16,0
	lsr	R16
	lsr	R16
	;lsr	R16
	andi	R16,0x07
	rcall	LCDwrData
	ld	R16,Y
	subi	R16,20
	brge	PC+1+1
	ldi	R16,0	
	st	Y+,R16
	dbnz	R24,LCDwrMODlnBGL1
	.endif
	ret	
	